(***********************************************************************************
 *
 *  Application SRTFS (Software Reuse Through Formal Specifications) 
 *
 *  File: comp_match.sml
 *
 *  Description: functions that implement the composite matching
 * 
 ***********************************************************************************)

structure CompMatch =
    struct 

	(* for the debug *)
	structure Err=Err;	

	exception not_match

	(* In: usr_post = the user post-categories
	 *     lib_post_i = the post-categories from the library initial method
	 *     lib_post_f = the post-categories from the library final method
	 *     fct = the incomplete functor
	 * Out: the functor or failure 
	 * Description: for each post-category from usr_post, it finds
	 *              the respective post-category from the 
	 *              composite categorical specification. 
         *              Then, it tries to match the two categories. *)
	fun match_post_catgs (usr_post:Catg.poscats) 
	    (lib_post_i:Catg.poscats) (lib_post_f:Catg.poscats) fct=
	    let
		(* identifies one antecedent from the lib_post_i that matches
		 the given user antecedent. Then, it gives the post-category 
		 induced by the antecedent identified *)
		fun get_post_i ant = 
		    case (List.find (fn x => 
				     FunCatg.equal_obj (#1 x) 
				     (Fct.Indz.obj fct ant)) 
			  lib_post_i) of
			SOME (ant,cons) => cons
		      | NONE => (debug "get_post_i: fail\n";raise not_match)
		(* identifies the antecedent from the lib_post_f that verifies
		 the given lib_post_i post-category. 
		 Then, it gives the post-category induced
		 by the antecedent identified *)
		fun get_post_f cons = 
		    case (List.find (fn x => 
				     FunCatg.verify cons (#1 x)) 
			  lib_post_f) of
			SOME (ant,cons) => cons
		      | NONE => (debug "get_post_f: fail\n";raise not_match)
		(* get the library post_category 
		 derived from one user antecedent *)
		fun get_post ant = (get_post_f (get_post_i ant))
		val _ = IsoMatch.t_ctg := IsoMatch.post
	    in    
		debug "match_post_catgs: start\n";
		(Tree.and_match (List.map (fn (a,c) => 
					   (c,(get_post a))) usr_post) 
		 IsoMatch.match_catg fct
		 handle not_match => NONE)
	    end    

	(* In: match = message that contains the name of the components
	 *     usr_mtd = the user method's categorical spec. 
	 *     lib_mtd_i = the library initial method's categorical spec. 
	 *     lib_mtd_f = the library final method's categorical spec. 
	 * Out: prints the functor if it identifies a match
	 * Description: it starts by identifying a isomorphic matching between
	 *              the usr_mtd pre-category and the lib_mtd_i pre-category.
	 *              Then, it'll match the post-categories. *)
	fun match_metd match (usr_mtd:Catg.method) 
	    (lib_mtd_i:Catg.method) (lib_mtd_f:Catg.method) =
	    let
		val fct = Fct.new ()
		val _ = Stack.new ()
		val _ = IsoMatch.t_ctg := IsoMatch.pre
		val _ = Stack.push (match_post_catgs (#3 usr_mtd) 
				    (#3 lib_mtd_i) (#3 lib_mtd_f))
		val match = (match^"\t mtd: " ^(#1 usr_mtd)^
			     " <-> "^(#1 lib_mtd_i)^":"^(#1 lib_mtd_f)^"\n");
	    in
		case IsoMatch.match_catg (#2 usr_mtd) (#2 lib_mtd_i) fct of
		    SOME(f)=> (print (match^"Solution:\n"); Fct.out print f)
		  | NONE => ()
	    end

	(* In: usr_cmp = the user component's categorical spec. 
	 *     lib_cmp = the library component's categorical spec. 
	 * Out: nothing
	 * Description: it applies the match_metd function to
	 *              all the usr_cmp methods combined with 
	 *              all possible pairs of methods from lib_cmp. *)
	fun match_comp (usr_cmp:Catg.component) (lib_cmp:Catg.component) = 
	    let
		val match = ("\nmatching comp: "^
			     (#1 usr_cmp)^" <-> "^(#1 lib_cmp)^"\n");
	    in
		List.app (fn u =>
			  (List.app (fn l_i =>
				     (List.app 
				      (match_metd match u l_i)
				      (#3 lib_cmp)))
			   (#3 lib_cmp)))
		(#3 usr_cmp)
	    end

    end









