(***********************************************************************************
 *
 *  Application SRTFS (Software Reuse Through Formal Specifications) 
 *
 *  File: fun_ctg.sml
 *
 *  Description: functions that manipulate the categorical specifications 
 * 
 ***********************************************************************************)

structure FunCatg =
    struct  
        
	exception non_composable_pair;
	structure Utils=Utils;
        structure Catg=Catg;

	(* get data from an arrow *)
	fun source((a,_))=a;
	fun target((_,b))=b;
	 
	(* returns the identity arrow of an object *)    
	fun ident(a)=(a,a);

	(* returns the terminal(initial) arrow of an object *)    
	fun arr_term(a)=(a,Catg.terminal); 
	fun arr_ini(a)=(Catg.initial,a); 

	(* returns a empty category *)
	fun empty () = ([],[])
	    
	(* classify the type of an assertion *)
	val ast_type =
	    fn Catg.prop(_,_) => 1 
	     | Catg.ext(_) => 2
	     | Catg.exp(_) => 3

	(* classify the type of an object *)
	val obj_type = 
	    fn 
	    Catg.terminal => 0 | Catg.initial => 0
	  | Catg.sing(_) => 1
	  | Catg.min(x) => ~(List.length x)
	  | Catg.max(x) => (List.length x)

	(* returns the assertions of an object *)
	val assertions = 
	    fn 
	    Catg.terminal => [] | Catg.initial => []
	  | Catg.sing(x) => [x]
	  | Catg.min(x) => x
	  | Catg.max(x) => x

	(* classify the type of a predicate *)
	val pred_type =
	    fn Catg.pred(_) => 1 
	     | Catg.comp(_) => 2

	(* returns the identifier of a predicate *)
	val pred_id = 
	    fn 
	    Catg.pred(id) => id
	  | Catg.comp(Catg.in_(id)) => id
	  | Catg.comp(Catg.out(id)) => id
	
	(* compare predicates, ignoring the type of variables *)
	fun cmp_pred p1 p2 =
	    p1=p2 orelse 
	    ((pred_type p1)=(pred_type p2) andalso (pred_id p1)=(pred_id p2))

	(* compare assertions, ignoring the type of variables *)
	fun cmp_assertion a1 a2 =
	    case a1 of 
		Catg.prop(p1,c1) => 
		    (case a2 of
			 Catg.prop(p2,c2) => (cmp_pred p1 p2) andalso c1=c2
		       | _ => false)
	      | _ => a1=a2

	(* compare objects *)    
	fun equal_obj x y = 
	    (obj_type x = obj_type y) andalso 
	    Utils.setcont ((assertions x), (assertions y))

	(* returns the composition of two arrows *)
	fun comp((c,d),(a,b))=
	    if equal_obj b c then (a,d)
	    else Err.error Err.non_composable_pair "";
		
	(* compare arrows *)
	fun equal_arr (xs,xd) (ys,yd) = 
	    (equal_obj xs ys) andalso (equal_obj xd yd)

	(* verifies if one arrow implies the other *)
        fun match_arr (xs,xd) (ys,yd) =
            let 
		val s = (Utils.setcont_ cmp_assertion ((assertions xs), 
						       (assertions ys)))
	            andalso (not((obj_type xs)=0) orelse (equal_obj xs ys))
		val d = (Utils.setcont_ cmp_assertion ((assertions yd), 
						       (assertions xd)))
		    andalso (not((obj_type yd)=0) orelse (equal_obj xd yd))
	    in
		s andalso d
	    end

	(* verifies if a category satisfies an object *)
	fun verify (catg:Catg.catg) obj =
	    let
		val objs = case obj of
		    (* if the object represent a conjunction then
		     it have to satisfy the objects induced
		     by its assertions *)
		    Catg.min(x) => (List.map 
				    (fn Catg.ext(y) => Catg.sing(Catg.exp(y))
				  | Catg.exp(y) => Catg.max([Catg.ext(y)])
				  | ast => Catg.sing(ast))
				    (assertions obj))
		  | _ => [obj]
		val arrs = List.map (fn obj => (Catg.terminal, obj)) objs
	    in
		Utils.setcont_ match_arr (arrs, (#2 catg))
	    end

	(* join two categories *)
	fun join (obj1,arr1) (obj2,arr2) =
	    (Utils.append obj1 obj2,
	     Utils.append arr1 arr2)

    end;










