(***********************************************************************************
 *
 *  Application SRTFS (Software Reuse Through Formal Specifications) 
 *
 *  File: iso_match.sml
 *
 *  Description: functions that implement the isomorphic matching
 * 
 ***********************************************************************************)

structure IsoMatch =
    struct 

	(* for the debug *)
	structure Err=Err;

	exception not_match;

	(* actual type of category match *)
	(* pre-category or post-category *)
	datatype type_ctg = pre | post; 
	val t_ctg = ref pre;
	
	(* In: list of objects, sorted by number of assertions 
         *     and type of object
	 * Out: a list of objects sub-lists
	 * Description: it creates a sub-list, for each set of objects
	 *              with the same number assertions and type *)
	val rec ord_obj = 
	    fn [] => []
	     | [x] => [[x]]
	     | h::t => 
	    case (ord_obj t) of
		r as hr::tr =>
		    if (FunCatg.obj_type (hd hr))=(FunCatg.obj_type h) then
			(h::hr)::tr
		    else [h]::r
	      | _ => Err.error Err.bug "IsoMatch.ord_obj"

	(* In: usr_obj = user object
	 *     lib_obj = library object
         *     fct = the actual functor
	 * Out: the functor or failure 
	 * Description: it tries to match the two objects. *)
	fun match_obj usr_obj lib_obj fct =
	    let
		(* verifies if it's possible to match the two
		 assertions given *)
		fun comb_p (usr_ast,lib_ast) = 
		    case (usr_ast,lib_ast) of
			(Catg.prop (pu,cu),Catg.prop (pl,cl)) =>
			    (((Fct.Indz.pred fct pu) = pl
			      handle Fct.Indz.cant_indz => true)
			     andalso
			     ((Fct.Indz.const fct cu) = cl
			      handle Fct.Indz.cant_indz => true))
		      | (Catg.exp (pu),Catg.exp (pl)) =>
			    ((Fct.Indz.pred fct pu) = pl
			     handle Fct.Indz.cant_indz => true)
		      | (Catg.ext (pu),Catg.ext (pl)) =>
			    ((Fct.Indz.pred fct pu) = pl
			     handle Fct.Indz.cant_indz => true)
		      | _ => false
		(* updates the functor with the match 
		 between the pair of assertions *)
		fun atrib pair fct = 
		    case pair of
			(Catg.prop(pu,cu), Catg.prop(pl,cl)) => 
			    (Fct.ins_pred 
			     (Fct.ins_const fct (cu,cl)) 
			     (pu,pl) 
			     )
		      | (Catg.ext(pu), Catg.ext(pl)) => 
			    (Fct.ins_pred fct (pu,pl))
		      | (Catg.exp(pu), Catg.exp(pl)) =>
			    (Fct.ins_pred fct (pu,pl))
		      | _ => Err.error Err.bug "IsoMatch.atrib"
		(* updates the functor with the match 
		 between the pairs of assertions *)
		fun match pairs fct =
		    ( debug "match_obj: found one possible solution\n";
		     Stack.call (Utils.nestlist atrib fct pairs)
		     handle Fct.bad_insert => NONE
			 )
	    in
		debug "match_obj: start\n";
		debug "\t"; Output.object_ debug usr_obj; debug " : ";
		Output.object_ debug lib_obj; 
		debug "\nActual Functor: "; Fct.out debug fct;
		Tree.or_match (FunCatg.assertions usr_obj) 
		(FunCatg.assertions lib_obj) 
		Fct.Indz.assertion (fn x => fn y => x=y) comb_p
	        match fct
	    end

	(* In: usr_objs = group of user objects
	 *     lib_objs = group of library objects
         *     fct = the actual functor
	 * Out: the functor or failure 
	 * Description: it tries to match the two groups of objects. *)
	fun match_objs usr_objs lib_objs fct =
	    let
		(* verifies if the two objects can be matched, 
		 comparing the type of their assertions *)
		fun comb_p (usr_obj,lib_obj) = 
		    let
			fun types obj = 
			    (List.map FunCatg.ast_type (FunCatg.assertions obj))
		    in
			(* the objects have the same number of assertions *)
			Utils.setcont (types usr_obj, types lib_obj)
		    end
			
	    in
		debug ("match objects: start\n");
		debug ("\ttype:"^Int.toString
		       (FunCatg.obj_type (hd usr_objs))^"\n");
		debug ("\t\t"^(Int.toString
			       (List.length usr_objs))^":"
		       ^(Int.toString(List.length lib_objs))^"\n");
		Tree.or_match usr_objs lib_objs Fct.Indz.obj 
		FunCatg.equal_obj comb_p 
		(fn c => Tree.and_match c match_obj) fct
	    end

	(* In: usr_objs = user objects
	 *     lib_objs = library objects
         *     fct = the actual functor
	 * Out: the functor or failure 
	 * Description: it tries to match all the objects. *)
	fun match_list_objs usr_objs lib_objs fct =
	    let
		(* combines each group of user objects with
		 the respective group of library object *)
		fun pairs usrs libs = 
		    case usrs of 
			[] => []
		      | u as hu::tu => 
			    case libs of
				[] => raise not_match 
			      | l as hl::tl => 
				    if (FunCatg.obj_type (hd hu))
					=(FunCatg.obj_type (hd hl)) 
					then if (List.length hu)<=(List.length hl)
						 then (hu,hl)::(pairs tu tl)
					     else raise not_match
				    else (pairs u tl)
						   
	    in    
		debug "match_list_objs: start\n";
		debug ("\tnumber of groups: "^(Int.toString
					       (List.length usr_objs))^":"
		       ^(Int.toString(List.length lib_objs))^"\n");
		(Tree.and_match (pairs (List.rev usr_objs) lib_objs) 
		 match_objs fct
		 handle not_match => 
		     (debug "match_list_objs: not_match\n";NONE))
	    end

	(* In: usr_arrs = user arrows
	 *     lib_arrs = library arrows
         *     fct = the actual functor
	 * Out: the functor or failure 
	 * Description: verifies if all the arrows can be matched. *)
	fun match_arrows usr_arrs lib_arrs fct =
	    let
		val _ = (debug 
			 ("match_arrows: start\n\tnumber of arrows: "
			  ^(Int.toString(List.length usr_arrs))^":"
			  ^(Int.toString(List.length lib_arrs))^"\n");
			 debug "\tuser arrows(translated): "; 
			 (List.app (Output.arrow debug)
			  (List.map (Fct.Indz.arrow fct) usr_arrs));
			 debug "\tlibrary arrows: ";
			 (List.app (Output.arrow debug) lib_arrs);
			 debug "\tActual Functor: ";
			 Fct.out debug fct)
		(* translates the user arrows using the functor *)
		val indz_usr_arrs = List.map (Fct.Indz.arrow fct) usr_arrs
		(* verifies if the arrows are from a 
		 pre-category or post-category *)
		val comp = if (!t_ctg)=post then (indz_usr_arrs, lib_arrs)
			   else (lib_arrs, indz_usr_arrs)

	    in
		if (Utils.setcont_ FunCatg.equal_arr comp)
		    then Stack.call fct
		else (debug "match_arrows: fail\n";NONE)
	    end

	(* In: usr_ctg = user category
	 *     lib_ctg = library category
         *     fct = the actual functor
	 * Out: the functor or failure 
	 * Description: it tries to match the two categories. *)
	fun match_catg (usr_ctg:Catg.catg) (lib_ctg:Catg.catg) fct = 
	    let
		val _ = debug "match_category: start\n"
		val usr_obj = ord_obj (#1 usr_ctg)
		val lib_obj = ord_obj (#1 lib_ctg)
		val _ = Stack.push (match_arrows (#2 usr_ctg) (#2 lib_ctg)) 
		val ret = match_list_objs usr_obj lib_obj fct
	    in
		(* cleans the stack *)
		Stack.pop (); Stack.pop (); 
		ret
	    end

	(* In: usr_post = user post-categories
	 *     lib_post = library post-categories
         *     fct = the actual functor
	 * Out: the functor or failure 
	 * Description: it tries to match the post-categories. *)
	fun match_post_catgs (usr_post:Catg.poscats) (lib_post:Catg.poscats) fct =
	    let
		val _ = debug "match_post_catgs: start\n"
		(* identifies the library post-category
		 that has an antecedent which matches with the
		 antecedent given *)
		fun get_post ant lib_post = 
		    case (List.find (fn x => 
				     FunCatg.equal_obj (#1 x) 
				     (Fct.Indz.obj fct ant)) 
			  lib_post) of
			SOME (ant,cons) => cons
		      | NONE => (debug "match_post_catgs: fail\n";raise not_match)
		val _ = t_ctg := post

	    in    
		(Tree.and_match (List.map (fn (a,c) => 
					   (c,get_post a lib_post)) usr_post) 
		 match_catg fct
		 handle not_match => NONE)
	    end    

	(* In: match = message that contains the name of the components
	 *     usr_mtd = user method
	 *     lib_mtd = library method
	 * Out: prints the functor if it identifies a match
	 * Description: it tries to match the two methods. *)
	fun match_metd match (usr_mtd:Catg.method) (lib_mtd:Catg.method) =
	    let
		val fct = Fct.new ()
		val _ = Stack.new ()
		val _ = t_ctg := pre
                val _ = Stack.push (match_post_catgs (#3 usr_mtd) (#3 lib_mtd))
		val match = (match^"\t mtd: "^
			     (#1 usr_mtd)^" <-> "^(#1 lib_mtd)^"\n")
	    in
		case match_catg (#2 usr_mtd) (#2 lib_mtd) fct of
		    SOME(f)=> (print (match^"Solution:\n"); Fct.out print f)
		  | NONE => ()
	    end

	(* In: usr_cmp = user component
	 *     lib_cmp = library component
	 * Out: nothing
	 * Description: it tries to match the two component's methods. *)
	fun match_comp (usr_cmp:Catg.component) (lib_cmp:Catg.component) = 
	    let
		val match = ("\nmatching comp: "^
			     (#1 usr_cmp)^" <-> "^(#1 lib_cmp)^"\n")
	    in
		List.app (fn u =>
			  (List.app
			   (match_metd match u)
			   (#3 lib_cmp)))
		(#3 usr_cmp)
	    end

    end







