(***********************************************************************************
 *
 *  Application SRTFS (Software Reuse Through Formal Specifications) 
 *
 *  File: transl.sml
 *
 *  Description: functions that translates an algebraic specification into 
 *               a categorical specification
 * 
 ***********************************************************************************)

structure Trans =
    struct
        
        exception semantic_error
	
	(* actual type of category translation *)
	(* pre-category or post-category *)
	datatype type_ctg = pre | post;
	val t_ctg = ref pre;

	(* predicate translation *)
	fun pred p =
	    case p of
		Spec.pred (id) => Catg.pred(id)
	      (* in the categorical there are no state variables *)
	      | Spec.comp (Spec.st(id)) => if (!t_ctg)=pre
					       then Catg.comp(Catg.in_(id))
					   else Catg.comp(Catg.out(id))
	      | Spec.comp (Spec.in_(id)) => Catg.comp(Catg.in_(id))
	      | Spec.comp (Spec.out(id)) => Catg.comp(Catg.out(id));
       
					       
	(* In: conj = indicates if the pseudo-proposition is part
	 *            of a conjunction or a disjunction
	 *     prop = a pseudo-proposition
	 * Out: an assertion 
	 * Description: translates a pseudo-proposition into an assertion *)
	fun proposition conj prop =
	    case prop of 
	        Spec.prop (p,term) =>
		    ( 
		     (Catg.prop (pred p,Output.term term)))
	      | Spec.all (p,x) => 
		    ( 
		     if conj then (Catg.ext (pred p))
		     else (Catg.exp (pred p))
			 )
	      | Spec.one (p,x) =>
		    ( 
		     if conj then (Catg.exp (pred p))
		     else (Catg.ext (pred p))
			 ) 
	      | _ => Err.error Err.bug 
		    "Trans.proposition, Terminal or initial out of place" 
		    
	(* In: sc = the antecedent or the consequent part of a clause 
	 * Out: an object 
	 * Description: translates the sub-clause into an object *)
	fun sub_clause sc =
	    let 
		(* conj = indicates if the pseudo-proposition is part
		 *        of a conjunction or a disjunction
		 * ps = the list of pseudo-propositions from sf *)
		val (conj, ps) =
		    case sc of
			Spec.conj (x) => (true,x)
		      | Spec.disj (x) => (false,x)
		(* creates the right type of object *)
		fun typ conj = 
		    (* one single object for each expansion *)
		    fn [Catg.exp(x)] => Catg.sing(Catg.exp(x))
		     (* only one proposition *)
		     | [Catg.prop(p,t)] => Catg.sing(Catg.prop(p,t)) 
		     | x => if conj then Catg.min(x)
			    else Catg.max(x)
		(* treats the special propositions, and 
		 creates the respective object *)
		fun object conj ps =
		    let 
			(* list that contains the special propositions and 
			 the objects induced by them *)
			val ord = (Spec.false,Catg.initial,
				   Spec.true,Catg.terminal)
			val ord = if conj then ord 
				  else (#3 ord,#4 ord,
					#1 ord,#2 ord)
			(* removes the neutral proposition *)
			val new_ps = (List.filter (fn x => not((#3 ord)=x)) ps)
		    in
			(* tries to find the absorbent proposition *)
			if (List.exists (fn x => ((#1 ord)=x)) new_ps) then
			    (#2 ord)
			else if new_ps=[] then
			    (#4 ord)
			     else typ conj (List.map (proposition conj) 
					    new_ps)
		    end
	    in
		object conj ps
	    end

	(* In: the antecedent and the consequent part of a clause 
	 * Out: one category 
	 * Description: translates the clause into an arrow and
	 *              its objects *)
	fun clause (Spec.prop_impl(antc,cons)) = 
	    let 
		val ant=sub_clause antc;
		val con=sub_clause cons;
	    in
		([ant,con],[(ant,con)]) 
	    end
	    
	(* In: a set of clauses 
	 * Out: one category
	 * Description: translates all the clauses into one category *)
	fun clausal_afirm ca = 
	    Utils.nestlist 
	    (fn cl => (FunCatg.join (clause cl)))  
	    (FunCatg.empty ())
	    ca

	(* In: list of objects
	 * Out: a list of objects sub-lists
	 * Description: it creates a sub-list, for each set of objects
	 *              with the same number assertions and type *)
	fun ord_obj objs =
	    let
		(* inserts the object x in the list res on its right place *) 
		fun ins x res =
		    case res of
			[] => [[x]]
		      | h::t => let (* type of objects *) 
				    val tx = (FunCatg.obj_type x)
				    val th = (FunCatg.obj_type (hd h))
				in
				    (* same type *)
				    if (tx=th) then 
					(x::h)::t
				    else 
					if (abs tx)<(abs th) then
					    [x]::res
					else h::(ins x t)
				end
	    in
		List.concat (Utils.nestlist ins [] objs)
	    end


	(* In: a = a list of clauses
	 * Out: a category 
	 * Description: translates the clauses a into a category *)
	fun trans_forms a =
	    (
	     case clausal_afirm a of
		 (obj,arr) => (ord_obj obj,arr)
		     )
	   
	(* In: name_comp = component's name
	 *     sort = single sort
	 *     variables = states variables
	 *     inv = invariant
	 *     name_metd = method's name
	 *     interface = interface variables
	 *     requires = pre-condition
	 *     ensures = post-condition
	 * Out: the categorical specification
	 * Description: translates the algebraic specification of a method
	 *              into its categorical specification *)
	fun trans_method (name_comp,sort,variables,inv) 
	    (name_metd,interface,requires,ensures) =
	    let
		(* gives the implications between all the post-conditions 
		 antecedents with the req *)
		fun comb req =
		    List.map
		    (fn (ant,cons) => Spec.prop_impl(ant,req))
		    ensures
		(* gives the implications between all the post-conditions 
		 antecedents with the pre-condition objects  *)
		val antecedents = 
		    List.concat (List.map comb requires)
		(* says that the pre-condition must be true *)
		val clausal_pre = 
		    List.map (fn r => 
			      Spec.prop_impl(Spec.conj([Spec.true]),r)) 
		    requires
		val _ = t_ctg := pre
		(* gets the pre-category *)
		val precat =
		    trans_forms (antecedents@clausal_pre@inv)
		(* gets the post-category *)
		fun poscat (ant,cons) =
		    let
			val _ = t_ctg := pre
			val ant_obj = (sub_clause ant)
			val _ = t_ctg := post
			val cons_ctg = trans_forms (cons@inv)
		    in 
			(ant_obj, cons_ctg)
		    end
		(* gets all the post-categories *)
		val poscats = List.map poscat ensures
	    in
		(* the categorical specification *)
		(name_metd, precat, poscats)
	    end;
	    
	(* In: name_comp = component's name
	 *     sort = single sort
	 *     variables = state variables
	 *     inv = invariant
	 *     methods = list of methods
	 * Out: a list of categorical specifications 
	 * Description: gives the list with the categorical specification
	 *              of each method *)
	fun trans_spec (name,sort,variables,inv,methods)=
	    (name,sort,
	     List.map 
	     (trans_method (name,sort,variables,inv)) 
	     methods)
	    
    end;       
    
    







    